#include "r2-asm.h"

TAB = 9

.text

proc ModuleXInit
	stmfd sp!, {r0-r4, lr}

	mov r0, #OS_PrettyPrint
	adr r1, xos_prettyprint
	mov r2, r12
	swi X|OS_ClaimSWI

	ldmfd sp!, {r0-r4, pc}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Take R0 as pointer to null-terminated string, print it
// with line breaks before words that won't fit on the current line.
// CR forces a newline. TAB=tab to next column of 8. CHR$31 = hard space
// <27,n> => substitute nth dictionary entry:
//  dictionary is pointed at by r1, consists of a list of null-terminated
//  strings, with preceding length bytes.
// 0th dictentry => use string at r2.

// Inside here, r1 is the current string pointer
//              R3 is the linelength to format to
//              R4 is the current position in the line.
//              r6 is the return addr in word counting
//              r7 is the r1 restore value in reading word length
//              r8 is used in dictionary selection in getbytepp
//              r9 is the sp restoration level for exiting.
//              r11 is the dictionary pointer (0 means use MOS dict)
//              r12 is the special token pointer

xos_prettyprint:
			STMFD   sp!, {r0-r9, lr}

			BL      ReadWindowWidth    		// read linelength
			MOV     R3, R0
			MOV     R0, #165          		 // read output cursor
			SWI     X|OS_Byte
			ORR     R4, R1, #0x80000000 		// no leading space even if in line.
			LDMFD   sp, {r1, r11, r12}   // reload strptr, dictptr, tokptr
			CMP     r11, #0
			ADREQL   r11, MOSdictionary
			MOV     r9, sp

// loop over all words
01:       BL      getwordlength
          CMP     R2, #0
          BNE     03f
          BL      getbytepp
          B       21f                // null word - test for done
03:

          CMP     R4, #0
          ADDGT   R2, R2, #1           // allow for separator
          ADD     R0, R2, R4
          BIC     R0, R0, #0x80000000   // clear R4 flag
          CMP     R0, R3
          BLE     10f
          CMP     R4, #0
          SUBGT   R2, R2, #1          // remove leading space length
          MOV     R4, #0              // if word too long to fit, do newline
          SWI     X|OS_NewLine
          BVS     exitpp

10:       CMP     R4, #0
          BIC     R4, R4, #0x80000000  // clear "no leading space" flag
          ADD     R4, R4, R2
          SUBGT   R2, R2, #1
          SWIGT   X|OS_WriteI|' '    // output separator if not 1st on line
          BVS     exitpp

04:       BL      getbytepp
          CMP     R0, #31           // hard space?
          MOVEQ   R0, #' '
          SWI     X|OS_WriteC
          BVS     exitpp
          SUBS    R2, R2, #1
          BNE     04b

21:       CMP     R0, #13
          MOVEQ   R4, #0
          SWIEQ   X|OS_NewLine
          BVS     exitpp

          CMP     R0, #TAB
          BEQ     20f
          CMP     R0, #0
          BNE     01b

exitpp:   MOV     sp, r9
          STRVS   r0, [sp]
          LDMFD   sp!, {r0-r9, pc}

// TAB had: align to next multiple of 8
20:       BIC     R4, R4, #0x80000000
   // first want to get next word length, to see if it's worth doing
          ADD     R5, R4, #8
          BIC     R5, R5, #7
          SUB     R5, R5, R4           // spaces for this tab

24:       BL      getwordlength
          CMP     R2, #0
          BNE     23f              // got the word
          BL      getbytepp
          CMP     R0, #13
          BEQ     21b              // TAB, CR - whalley!
          CMP     R0, #TAB
          SUBNE   R5, R5, #1         // leading spaces, junk ignored
          ADDEQ   R5, R5, #8         // multiple tabs OK tho.
          CMP     r0, #0             // terminator?
          BNE     24b              // only case to allow zero-length word
          SUB     r1, r1, #1         // we know this rewinds OK

23:       ADD     R0, R4, R5
          ADD     R0, R0, R2
          CMP     R0, R3
          MOVGT   R0, #13           // next tab stop too far : newline
          BGT     21b
22:       SWI     X|OS_WriteI|' '
          BVS     exitpp
          ADD     R4, R4, #1
          SUBS    r5, r5, #1
          BNE     22b
          ORR     R4, R4, #0x80000000 // set top bit to disable leading space
          B       01b

getwordlength:
          MOV     r6, lr
          MOV     r10, r9            // first copy context
          MOV     r2, sp
          MOV     r7, r1
copycontextpp:
          CMP     r9, r2
          LDRNE   r0, [r9, #-4]!
          STRNE   r0, [sp, #-4]!
          BNE     copycontextpp

          MOV     r2, #0             // word length
02:       BL      getbytepp
          CMP     R0, #31
          CMPNE   R0, #' '+1
          ADDGE   r2, r2, #1
          BGE     02b
          MOV     r1, r7
          MOV     sp, r9
          MOV     r9, r10
          MOV     pc, r6

getbytepp:LDRB    r0, [r1], #1
          CMP     r0, #TokenEscapeChar
          BEQ     gettokenpp
          CMP     r0, #0
          MOVNE   pc, lr
          CMP     sp, r9
          MOVHS   pc, lr
          LDR     r1, [sp], #4      // back to previous token
          B       getbytepp

gettokenpp:
          LDRB    r0, [r1], #1      // tokno
          LDR     r1, [sp], #4      // save context
          CMP     r0, #0
          MOVEQ   r1, r12
          BEQ     getbytepp
          MOV     r1, r11
gtoklp:   SUBS    r0, r0, #1
          LDRNEB  r8, [r1]
          ADDNE   r1, r1, r8
          BNE     gtoklp
          ADD     r1, r1, #1
          B       getbytepp

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ReadWindowWidth:    // read "number of chars printable after Newline
                    //       before we're on the next line"
			MOV     R0, #VduExt_WindowWidth
			STMFD   sp!, {r0, r1, lr}
			MOV     R0, #-1
			STR     R0, [sp, #4]  // overwrite R1 value
			MOV     R0, sp
			MOV     R1, sp
			SWI     X|OS_ReadVduVariables
			ldmfd sp!, {r0, lr, pc}

proc ModuleTitle
	.ascii "R4PrettyPrint"
	.byte 0

proc ModuleHelp
	.ascii "R4 PrettyPrinter"
	.byte 9
	.ascii "0.1 (23 September 2008)"
	.byte 0


MOSdictionary:

.macro TokenString n, s
0:
	.byte 1f - 0b
	.asciz "\s"
1:
.endm

	TokenString 1, "Syntax: *\x1b\x00"
	TokenString 2, " the "
	TokenString 3, "director"
	TokenString 4, "filing system"
	TokenString 5, "current"
	TokenString 6, " to a variable. Other types of value can be assigned with *"
	TokenString 7, "file"
	TokenString 8, "default "
	TokenString 9, "tion"
	TokenString 10, "*Configure "
	TokenString 11, "name"
	TokenString 12, " server"
	TokenString 13, "number"
	TokenString 14, "\x1b\x01 <"
	TokenString 15, " one or more \x1b\x07s that match\x1b\x02given wildcard"
	TokenString 16, " and "
	TokenString 17, "relocatable module"
	TokenString 18, "\rC(onfirm)\tPrompt for confirma\x1b\x09 of each "
	TokenString 19, "sets\x1b\x02"
	TokenString 20, "\x1b\x01 [<disc spec.>]"
	TokenString 21, "}\rV(erbose)\tPrint informa\x1b\x09 on each \x1b\x07 "
	TokenString 22, ""
	TokenString 23, "\x1b\x1fLandscape [<XScale> [<YScale> [<Margin> [<Threshold>]]]]]"
	TokenString 24, "\x1b\x29used\x1b\x28print a hard copy of\x1b\x02screen on EPSON-"
	TokenString 25, ".\rOp\x1b\x09s: (use ~\x1b\x28force off, eg. ~"
	TokenString 26, "printe"
	TokenString 27, "\x1b\x0e\x1b\x07\x1b\x0b>"
	TokenString 28, " select"
	TokenString 29, "xpression"
	TokenString 30, "\x1b\x01 ["
	TokenString 31, "sprite"
	TokenString 32, " displays"
	TokenString 33, "free space"
	TokenString 34, " {off}"
	TokenString 35, "library"
	TokenString 36, "parameter"
	TokenString 37, "object"
	TokenString 38, " all "
	TokenString 39, "disc"
	TokenString 40, " to "
	TokenString 41, " is "
	.byte 0
